import os
from src.sapphirerenderer.utility_objects.camera import Camera
import numpy as np
from src.sapphirerenderer.settings import (
    camera_move_speed,
    camera_rotate_speed,
    fps,
    show_fps,
    lock_fps,
)
from time import time
import threading


class SapphireRenderer:
    def __init__(self, width=1000, height=1000, draw_axis=False, object_files=None):
        """
        Initialize the renderer
        :param width: Width of the window
        :param height: Height of the window
        :param draw_axis: Draws the axis lines, use-full for debugging
        ;param: object_files: Used for loading different objects, a list of python file paths
        """
        self.display = None

        self.width = width
        self.height = height

        self.camera = Camera(self, position=np.array((0.0, -3.0, 0.0)))

        self.loaded_objects = []
        self.instance_objects = []
        self.load_objects()

        if draw_axis:
            self.add_object("Axes")

        self.running = True

        self.thread = threading.Thread(target=self.render_loop)
        self.thread.start()

    def load_objects(self):
        # go through all files in objects and load them
        for file in os.listdir("objects"):
            if file.endswith(".py"):
                exec(f"from objects.{file[:-3]} import *")
                obj_class_name = (
                    f"{file[:1].upper().replace('_', '')}{file[1:-3].replace('_', '')}"
                )
                self.loaded_objects.append((obj_class_name, eval(obj_class_name)))

    def add_object(self, obj_name, args=None):
        """
        Adds an object to the scene
        :param obj_name: The class name of the object
        :param args: The args to pass to the init of the class
        :return: returns the object created
        """
        for obj_class_name, obj_class in self.loaded_objects:
            if obj_class_name == obj_name:
                obj = obj_class(*args) if args is not None else obj_class()
                self.instance_objects.append(obj)
                return obj

    def update(self):
        self.camera.update()
        for obj in self.instance_objects:
            obj.update()

    def user_input(self, pygame, scale_factor=1.0):
        # wasd to move camera
        keys = pygame.key.get_pressed()
        if keys[pygame.K_w]:
            self.camera.move_relative((camera_move_speed * scale_factor, 0, 0))
        if keys[pygame.K_s]:
            self.camera.move_relative((-camera_move_speed * scale_factor, 0, 0))
        if keys[pygame.K_a]:
            self.camera.move_relative((0, camera_move_speed * scale_factor, 0))
        if keys[pygame.K_d]:
            self.camera.move_relative((0, -camera_move_speed * scale_factor, 0))
        if keys[pygame.K_q]:
            self.camera.move_relative((0, 0, -camera_move_speed * scale_factor))
        if keys[pygame.K_e]:
            self.camera.move_relative((0, 0, camera_move_speed * scale_factor))

        if keys[pygame.K_LEFT]:
            self.camera.rotate_relative((0, -camera_rotate_speed * scale_factor))
        if keys[pygame.K_RIGHT]:
            self.camera.rotate_relative((0, camera_rotate_speed * scale_factor))
        if keys[pygame.K_UP]:
            self.camera.rotate_relative((-camera_rotate_speed * scale_factor, 0))
        if keys[pygame.K_DOWN]:
            self.camera.rotate_relative((camera_rotate_speed * scale_factor, 0))

    def render_loop(self):
        import pygame

        self.display = pygame.display.set_mode((self.width, self.height))
        self.display.fill((255, 255, 255))
        pygame.display.set_caption("Sapphire Renderer")

        while self.running:
            frame_start = time() + 0.00001

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.running = False

            self.display.fill((255, 255, 255))
            self.update()

            for obj in self.instance_objects:
                obj.draw(self.display, self.camera)

            pygame.display.flip()

            # if fps is higher than fps setting, wait
            if lock_fps and time() - frame_start < 1 / fps:
                pygame.time.wait(int(1000 * (1 / fps - (time() - frame_start))))

            real_fps = 1 / (time() - frame_start)

            if show_fps:
                pygame.display.set_caption(f"Sapphire Renderer - FPS: {int(real_fps)}")

            self.user_input(pygame, fps / real_fps)

        pygame.quit()

    def stop(self):
        self.running = False
        self.thread.join()



import numpy as np


def get_pitch_matrix(yaw):
    return np.array(
        [[np.cos(yaw), 0, np.sin(yaw)], [0, 1, 0], [-np.sin(yaw), 0, np.cos(yaw)]]
    )


def get_roll_matrix(pitch):
    return np.array(
        [
            [1, 0, 0],
            [0, np.cos(pitch), -np.sin(pitch)],
            [0, np.sin(pitch), np.cos(pitch)],
        ]
    )


def get_yaw_matrix(roll):
    return np.array(
        [[np.cos(roll), -np.sin(roll), 0], [np.sin(roll), np.cos(roll), 0], [0, 0, 1]]
    )


def get_pitch_yaw_matrix(pitch, yaw):
    return np.dot(get_roll_matrix(pitch), get_yaw_matrix(yaw))


def get_pitch_yaw_roll_matrix(yaw, pitch, roll):
    return np.dot(get_pitch_yaw_matrix(yaw, pitch), get_roll_matrix(roll))



from src.sapphirerenderer.object_classes.vert_line_object import VertLineObject
import numpy as np


class Cube(VertLineObject):
    def __init__(self, position=np.array([0.0, 0.0, 0.0]), color=(0, 0, 0), size=1):
        vertices = np.array(
            [
                np.array([0, 0, 0]),
                np.array([size, 0, 0]),
                np.array([size, size, 0]),
                np.array([0, size, 0]),
                np.array([0, 0, size]),
                np.array([size, 0, size]),
                np.array([size, size, size]),
                np.array([0, size, size]),
            ],
            dtype=float,
        )

        lines = [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 0],
            [4, 5],
            [5, 6],
            [6, 7],
            [7, 4],
            [0, 4],
            [1, 5],
            [2, 6],
            [3, 7],
        ]

        super().__init__(vertices, lines, position, color)

        self.move_relative(np.array([-size / 2, -size / 2, -size / 2]))



import numpy as np
from stl import mesh
from src.sapphirerenderer.object_classes.vert_line_object import VertLineObject


class Stl(VertLineObject):
    def __init__(self, filename, position=np.array([0.0, 0.0, 0.0]), color=(0, 0, 0)):
        # Load STL file
        mesh_data = mesh.Mesh.from_file(filename)

        # Extract vertices
        vertices = np.unique(mesh_data.vectors.reshape(-1, 3), axis=0)

        # Generate line segments
        line_segments = []
        for triangle in mesh_data.vectors:
            for i in range(3):
                p1 = np.nonzero((vertices == triangle[i]).all(axis=1))[0][0]
                p2 = np.nonzero((vertices == triangle[(i + 1) % 3]).all(axis=1))[0][0]
                line_segments.append((p1, p2))

        super().__init__(vertices, np.array(line_segments), position, color)

    def update(self):
        self.rotate(0.01, 0.01, 0.01)



from src.sapphirerenderer.object_classes.base_object import Object
import numpy as np
import pygame
from src.sapphirerenderer.settings import (
    draw_vertices,
    draw_lines,
    line_thickness,
    point_thickness,
)
from src.sapphirerenderer.point_math.project_point import project_point
from src.sapphirerenderer.point_math.matricies import get_pitch_yaw_roll_matrix


class VertLineObject(Object):
    def __init__(
        self,
        vertices,
        lines,
        position=np.array([0, 0, 0], dtype=float),
        color=(0, 0, 0),
    ):
        super().__init__(position, color)
        self.original_vertices = vertices
        self.vertices = vertices
        self.lines = lines
        self.position = np.array([0, 0, 0], dtype=float)
        self.color = color

        self.move_absolute(position)

    def move_relative(self, vector):
        """
        Move the object by a relative amount
        :param vector: the amount to move by
        :return:
        """
        self.position += vector
        for i in range(len(self.vertices)):
            self.vertices[i] += vector

    def move_absolute(self, vector):
        """
        Move the object to an absolute position
        :param vector: the position to move to
        :return:
        """
        self.position = vector
        for i in range(len(self.vertices)):
            self.vertices[i] = self.original_vertices[i] + vector

    def rotate(self, pitch, yaw, roll):
        rotation_matrix = get_pitch_yaw_roll_matrix(pitch, yaw, roll)
        self.vertices = np.dot(self.vertices, rotation_matrix.T)

    def __str__(self):
        return self.__class__.__name__

    def draw(self, surface, camera):
        if self.lines is not None:
            for line in self.lines:
                start = self.vertices[line[0]]
                end = self.vertices[line[1]]

                start, s_scale = project_point(
                    start,
                    camera.rotation_matrix,
                    camera.position,
                    camera.size,
                    camera.focal_length,
                )
                end, e_scale = project_point(
                    end,
                    camera.rotation_matrix,
                    camera.position,
                    camera.size,
                    camera.focal_length,
                )

                if draw_vertices:
                    if start is not None:
                        pygame.draw.circle(
                            surface,
                            self.color,
                            start,
                            max(int(point_thickness * s_scale), 1),
                        )
                    if end is not None:
                        pygame.draw.circle(
                            surface,
                            self.color,
                            end,
                            max(int(point_thickness * e_scale), 1),
                        )

                if start is None or end is None:
                    continue

                if draw_lines:
                    pygame.draw.line(
                        surface,
                        line[2] if len(line) > 2 else self.color,
                        start,
                        end,
                        max(int(line_thickness * (s_scale + e_scale) / 2), 1),
                    )
        else:
            for vertex in self.vertices:
                vertex, scale = project_point(
                    vertex,
                    camera.rotation_matrix,
                    camera.position,
                    camera.size,
                    camera.focal_length,
                )

                if draw_vertices and vertex is not None:
                    pygame.draw.circle(
                        surface,
                        self.color,
                        vertex,
                        max(int(point_thickness * scale), 1),
                    )



import numpy as np
from src.sapphirerenderer.point_math.matricies import get_pitch_yaw_matrix


class Camera:
    def __init__(
        self,
        renderer,
        position=np.array([0.0, 0.0, 0.0]),
        rotation=np.array((0.0, 0.0)),
        fov=80,
    ):
        self.position = position
        self.rotation = rotation
        self.fov = fov
        self.rotation_matrix = self.get_rotation_matrix()
        self.size = (renderer.width, renderer.height)
        self.focal_length = (self.size[0] / 2) / np.tan((self.fov / 2) * (np.pi / 180))

    def move_absolute(self, position):
        position = np.array([-position[1], position[0], position[2]])
        self.position = position

    def move_relative(self, position):
        position = np.array([-position[1], position[0], position[2]])
        self.position += position @ self.rotation_matrix

    def rotate_absolute(self, rotation):
        self.rotation = rotation

    def rotate_relative(self, rotation):
        self.rotation += rotation

    def update(self):
        self.rotation_matrix = self.get_rotation_matrix()

    def get_rotation_matrix(self):
        return get_pitch_yaw_matrix(*self.rotation)



import numpy as np


class Object:
    def __init__(self, position=np.array([0, 0, 0]), rotation=np.array((0, 0, 0)), scale=1, color=(0, 0, 0)):
        self.position = position
        self.rotation = rotation
        self.scale = scale
        self.color = color

    def set_position(self, position):
        self.position = position

    def set_rotation(self, rotation):
        self.rotation = rotation

    def set_scale(self, scale):
        self.scale = scale

    def set_color(self, color):
        self.color = color

    def get_position(self):
        return self.position

    def get_rotation(self):
        return self.rotation

    def get_scale(self):
        return self.scale

    def get_color(self):
        return self.color

    def update(self):
        """
        empty update for child classes to override
        :return:
        """
        pass

    def __str__(self):
        return self.__class__.__name__



from src.sapphirerenderer.object_classes.vert_line_object import VertLineObject
import numpy as np


class Axes(VertLineObject):
    def __init__(self, position=np.array([0.0, 0.0, 0.0]), color=(0, 0, 0)):
        vertices = np.array(
            [
                np.array([0, 0, 0]),
                np.array([1, 0, 0]),
                np.array([0, 1, 0]),
                np.array([0, 0, 1]),
            ],
            dtype=float,
        )

        lines = [
            [0, 1, (159, 0, 0)],
            [0, 2, (0, 159, 0)],
            [0, 3, (0, 0, 159)],
        ]

        super().__init__(vertices, lines, position, color)



import numpy as np
from src.sapphirerenderer.object_classes.vert_line_object import VertLineObject


def generate_pyramid_points_and_segments(base_side_length, height):
    # Define the base vertices
    base_vertices = np.array(
        [
            [-base_side_length / 2, -base_side_length / 2, 0],
            [-base_side_length / 2, base_side_length / 2, 0],
            [base_side_length / 2, base_side_length / 2, 0],
            [base_side_length / 2, -base_side_length / 2, 0],
        ]
    )

    # Define the apex vertex
    apex_vertex = np.array([0, 0, height])

    # Combine vertices
    vertices = np.vstack((base_vertices, apex_vertex))

    # Define base lines
    base_lines = [[0, 1], [1, 2], [2, 3], [3, 0]]

    # Define lines from apex to base vertices
    apex_lines = [[4, i] for i in range(4)]

    # Combine lines
    lines = np.vstack((base_lines, apex_lines))

    return vertices, lines


class Pyramid(VertLineObject):
    def __init__(
        self,
        base_side_length=1.0,
        height=1.0,
        position=np.array([0.0, 0.0, 0.0]),
        color=(0, 0, 0),
    ):
        vertices, lines = generate_pyramid_points_and_segments(base_side_length, height)
        super().__init__(vertices, lines, position, color)



import numpy as np
from numba import njit


@njit(fastmath=True)
def project_point(point, rotation_matrix, camera_position, camera_size, focal_length):
    rotated_point = rotation_matrix @ (point - camera_position)
    y_rotated = rotated_point[1]

    if y_rotated < 0:
        return None, 1

    focal_length_divided_by_y = focal_length / y_rotated

    projected_point = np.array([
        rotated_point[0] * focal_length_divided_by_y,
        -rotated_point[2] * focal_length_divided_by_y,
    ])

    offset_x = camera_size[0] / 2
    offset_y = camera_size[1] / 2

    offset_array = np.array([offset_x, offset_y])

    projected_point += offset_array

    scale_factor = 1 / np.linalg.norm(rotated_point)

    return projected_point, scale_factor



import numpy as np
from src.sapphirerenderer.object_classes.vert_line_object import VertLineObject


def generate_torus_points_and_segments(radius_major, radius_minor, resolution):
    u = np.linspace(0, 2 * np.pi, resolution)
    v = np.linspace(0, 2 * np.pi, resolution)
    U, V = np.meshgrid(u, v)

    x = (radius_major + radius_minor * np.cos(V)) * np.cos(U)
    y = (radius_major + radius_minor * np.cos(V)) * np.sin(U)
    z = radius_minor * np.sin(V)

    points = np.stack((x.flatten(), y.flatten(), z.flatten()), axis=-1)

    # Generate line segments
    segments = []
    for i in range(resolution):
        for j in range(resolution):
            index = i * resolution + j
            if j < resolution - 1:
                segments.append([index, index + 1])
            else:
                segments.append([index, i * resolution])

            if i < resolution - 1:
                segments.append([index, index + resolution])
            else:
                segments.append([index, j])

    return points, np.array(segments)


class Torus(VertLineObject):
    def __init__(self, position=np.array([0.0, 0.0, 0.0]), color=(0, 0, 0)):
        vertices, lines = generate_torus_points_and_segments(2, 0.3, 20)

        super().__init__(vertices, lines, position, color)



import numpy as np
from src.sapphirerenderer.object_classes.vert_line_object import VertLineObject


def generate_sphere_points_and_segments(radius, resolution):
    theta = np.linspace(0, np.pi, resolution)
    phi = np.linspace(0, 2 * np.pi, resolution)
    THETA, PHI = np.meshgrid(theta, phi)

    x = radius * np.sin(THETA) * np.cos(PHI)
    y = radius * np.sin(THETA) * np.sin(PHI)
    z = radius * np.cos(THETA)

    points = np.stack((x.flatten(), y.flatten(), z.flatten()), axis=-1)

    # Generate line segments
    segments = []
    for i in range(resolution):
        for j in range(resolution):
            index = i * resolution + j
            if j < resolution - 1:
                segments.append([index, index + 1])
            else:
                segments.append([index, i * resolution])

            if i < resolution - 1:
                segments.append([index, index + resolution])
            else:
                segments.append([index, j])

    return points, np.array(segments)


class Sphere(VertLineObject):
    def __init__(
        self,
        radius=1.0,
        position=np.array([0.0, 0.0, 0.0]),
        color=(0, 0, 0),
        resolution=20,
    ):
        vertices, lines = generate_sphere_points_and_segments(radius, resolution)
        super().__init__(vertices, lines, position, color)
